@@ -4,7 +4,7 @@ use std::cell::Cell;
 use std::path::Path;
 use std::rc::Rc;
 
-use image::{EncodableLayout, Rgba, RgbaImage, SubImage};
+use image::{Rgba, RgbaImage, SubImage};
 
 use crate::error::{Result, TetraError};
 use crate::fs;
@@ -49,13 +49,12 @@ impl PartialEq for TextureSharedData {
 ///
 /// # Performance
 ///
-/// Creating a `Texture` is a relatively expensive operation. If you can, store them in your
-/// [`State`](crate::State) struct rather than recreating them each frame.
+/// Creating a texture is quite an expensive operation, as it involves 'uploading' the texture
+/// data to the GPU. Try to reuse textures, rather than recreating them every frame.
 ///
-/// Cloning a `Texture` is a very cheap operation, as the underlying data is shared between the
-/// original instance and the clone via [reference-counting](https://doc.rust-lang.org/std/rc/struct.Rc.html).
-/// This does mean, however, that updating a `Texture` (for example, changing its filter mode) will also
-/// update any other clones of that `Texture`.
+/// You can clone a texture cheaply, as it is a [reference-counted](https://doc.rust-lang.org/std/rc/struct.Rc.html)
+/// handle to a GPU resource. However, this does mean that modifying a texture (e.g.
+/// setting the filter mode) will also affect any clones that exist of it.
 ///
 /// # Examples
 ///
@@ -154,7 +153,7 @@ impl Texture {
     ) -> Result<Texture> {
         let handle = device.new_texture(width, height, filter_mode)?;
 
-        device.set_texture_data(&handle, &data, 0, 0, width, height)?;
+        device.set_texture_data(&handle, data, 0, 0, width, height)?;
 
         Ok(Texture {
             data: Rc::new(TextureSharedData {
@@ -357,7 +356,7 @@ impl Texture {
         data: &[u8],
     ) -> Result {
         ctx.device
-            .set_texture_data(&self.data.handle, &data, x, y, width, height)
+            .set_texture_data(&self.data.handle, data, x, y, width, height)
     }
 
     /// Overwrites the entire texture with new RGBA pixel data.
@@ -462,11 +461,6 @@ impl NineSlice {
 /// | ICO | `texture_ico` | No |
 /// | PNM | `texture_pnm` | No |
 /// | DDS/DXT | `texture_dds` | No |
-///
-/// # Performance
-///
-/// Creating or cloning an `ImageData` is a relatively expensive operation. If you can, store them in your
-/// [`State`](crate::State) struct rather than recreating them each frame.
 #[derive(Debug, Clone)]
 pub struct ImageData {
     data: RgbaImage,
@@ -570,7 +564,16 @@ impl ImageData {
 
     /// Returns the image's data, as a slice of raw bytes.
     pub fn as_bytes(&self) -> &[u8] {
-        self.data.as_bytes()
+        &self.data
+    }
+
+    /// Returns the image's data, as a mutable slice of raw bytes.
+    ///
+    /// This is not currently exposed publicly, as some more thought is needed
+    /// into whether doing so would cause issues once different pixel formats
+    /// are supported.
+    pub(crate) fn as_mut_bytes(&mut self) -> &mut [u8] {
+        &mut self.data
     }
 
     /// Returns the image's underlying buffer.
